import bpy
import bmesh
import numpy as np
import bgl
import blf
import gpu
import random
import mathutils
import os

from bpy.props import StringProperty, FloatVectorProperty, BoolProperty, EnumProperty, FloatProperty
from bpy.types import Operator
from mathutils import Color, Vector
from bpy_extras import mesh_utils
from gpu_extras.batch import batch_for_shader

from . visualuv_shaders import shader_info, shader_info_2D, shader_info_texture, shader_info_wireframe

def toggle_operation(context, operation):
    active_obj = context.object
    for obj in context.selected_objects:
        context.view_layer.objects.active = obj
        if obj.uv_operation == 'NONE' and not obj.uv_checker_texture:
            obj.uv_operation = operation
            obj.rv_enabled = True
            obj.data.update()
            bpy.ops.visualuv.overlay('INVOKE_DEFAULT')
        elif obj.uv_operation == operation and obj.uv_checker_texture:
            obj.uv_operation = 'NONE'
            obj.uv_recalculate = True
        elif obj.uv_operation != operation:
            obj.uv_operation = operation
            obj.rv_enabled = True
            obj.uv_recalculate = True
        else:
            obj.uv_operation = 'NONE'
            obj.rv_enabled = False
    context.view_layer.objects.active = active_obj

class VISUALUV_OT_update(Operator):
    bl_idname = "visualuv.update"
    bl_label = "Recalculate"
    bl_description = "Recalulates current operation"
    def execute(self, context):
        for obj in context.selected_objects:
            if obj.rv_enabled:
                obj.uv_recalculate = True
                if obj.uv_operation == 'UV_OVERLAP':
                    context.window_manager.uv_select_overlap = True
        return {'FINISHED'}

class VISUALUV_OT_toggle_texture(Operator):
    bl_idname = "visualuv.toggle_texture"
    bl_label = "Toggle texture rendering"
    bl_description = "Enable/Disable tile texture rendering"
    def execute(self, context):
        obj = context.object
        if not obj.rv_enabled:
            obj.uv_checker_texture = True
            obj.rv_enabled = True
            obj.data.update()
            bpy.ops.visualuv.overlay('INVOKE_DEFAULT')
        elif obj.uv_operation != 'NONE':
            obj.uv_checker_texture = not obj.uv_checker_texture
        else:
            obj.uv_checker_texture = False
            obj.rv_enabled = False
        return {'FINISHED'}

class VISUALUV_OT_toggle_stretching(Operator):
    bl_idname = "visualuv.toggle_stretching"
    bl_label = "Toggle stretching visualisation"
    bl_description = "Visualize stretching of uv faces"
    def execute(self, context):
        toggle_operation(context, 'UV_STRETCHING')
        return {'FINISHED'}


class VISUALUV_OT_toggle_normals(Operator):
    bl_idname = "visualuv.toggle_normals"
    bl_label = "Toggle uv normals visualisation"
    bl_description = "Visualize uv normals"
    def execute(self, context):
        toggle_operation(context, 'UV_NORMALS')
        return {'FINISHED'}

class VISUALUV_OT_toggle_overlapped(Operator):
    bl_idname = "visualuv.toggle_overlapped"
    bl_label = "Toggle overlapped uv visualisation"
    bl_description = "Visualize uv normals"
    def execute(self, context):
        toggle_operation(context, 'UV_OVERLAP')
        if context.object.uv_operation == 'UV_OVERLAP':
            context.window_manager.uv_select_overlap = True
        return {'FINISHED'}

class VISUALUV_OT_toggle_islands(Operator):
    bl_idname = "visualuv.toggle_islands"
    bl_label = "Toggle uv islands coloring"
    bl_description = "Color individual uv islands"
    def execute(self, context):
        toggle_operation(context, 'UV_ISLANDS')
        return {'FINISHED'}


class VISUALUV_OT_overlay(Operator):
    bl_idname = "visualuv.overlay"
    bl_label = "Visual UV face overlay operator"
    bl_description = "Draw Visual UV face overlay"

    def get_smallest_vector_dimension(self, vector):
        smallest_dimension = vector[0]

        for dimension in vector:
            if dimension < smallest_dimension:
                smallest_dimension = dimension

        return smallest_dimension

    def prep_wireframe_batch(self, shader, mesh, obj, vert_idx_cache, edge_indices):
        coords = np.empty((len(mesh.vertices), 3), 'f')
        mesh.vertices.foreach_get("co", np.reshape(coords, len(mesh.vertices) * 3))

        # offset wireframe verts to have slightly different depth - works well only for close view range
        for c_idx, coord in enumerate(coords):
            coords[c_idx] = coord + mesh.vertices[c_idx].normal*0.0035

        wireframe_colors = np.empty((len(mesh.vertices), 4), 'f')
        for v_idx, _ in enumerate(mesh.vertices):
            wireframe_colors[v_idx] = (0, 0, 0, obj.rv_groups_alpha) if v_idx in vert_idx_cache else (0, 0, 0, 0)

        return batch_for_shader(shader, 'LINES', {"position": coords, "color": wireframe_colors}, indices=edge_indices)

    def prep_pole_batch(self, shader, mesh, obj):
        if obj.mode == 'EDIT':
            bm = bmesh.from_edit_mesh(obj.data)
        else:
            bm = bmesh.new()
            bm.from_mesh(obj.data)

        bm.verts.ensure_lookup_table()
        bm.edges.ensure_lookup_table()
        bm.faces.ensure_lookup_table()

        pole_verts = [vert for vert in bm.verts if len(vert.link_edges) > 4]
        pole_coords = []
        pole_indices = []

        pole_idx = 0
        for vert in pole_verts:
            pole_coords.append(vert.co)

            smallest_dimension = self.get_smallest_vector_dimension(obj.dimensions)
            pole_coords.append(vert.co + vert.normal * smallest_dimension * 0.5 * obj.rv_poles_size)
            pole_indices.append([pole_idx, pole_idx+1])

            pole_idx += 2

        pole_color = (obj.rv_poles_color.r, obj.rv_poles_color.g, obj.rv_poles_color.b, 1)
        pole_colors = [pole_color for _ in pole_coords]

        return batch_for_shader(shader, 'LINES', {"position": pole_coords, "color": pole_colors}, indices=pole_indices)

    def recalculate_poly_islands(self, context, mesh):
        self.lst = mesh_utils.mesh_linked_uv_islands(mesh)
        self.directions = dict()
        self.island_colors = dict()
        color_value = 0.0
        step = 1.0 / len(self.lst)
        object_location = Vector(self.invoked_obj.location)
        for island in self.lst:
            island_center = Vector()
            color_value += step
            for poly in island:
                island_center += Vector(mesh.polygons[poly].center)
                self.island_colors[poly] = color_value
            island_center = island_center / len(island)
            if not self.invoked_obj.uv_enable_explosion_view:
                continue
            direction = island_center - object_location
            if direction == Vector():
                direction = Vector((0.0, 0.0, 1.0))
            else:
                direction = direction.normalized()
            for poly in island:
                self.directions[poly] = direction
 
    def compute_normal(self, vert1: Vector, vert2: Vector, vert3: Vector):
        edge1 = vert2 - vert1
        edge2 = vert3 - vert2
        return edge1.cross(edge2)

    def recalculate_uv_normals(self, obj, mesh):
        self.calc_colors = dict()
        uv_coords = [(0.0, 0.0, 0.0) for i in range(3)]

        for _, triangle in enumerate(mesh.loop_triangles):
            if mesh.polygons[triangle.polygon_index].hide and obj.mode == 'EDIT':
                continue

            for i in range(3):
                loop_index = triangle.loops[i]
                uv_coords[i] = Vector((*mesh.uv_layers.active.data[loop_index].uv, 0))
            uv_normal = self.compute_normal(uv_coords[0], uv_coords[1], uv_coords[2])

            for i in range(3):
                blue = 2.0 / 3.0 #blue
                red = 1.0
                key = (mesh.vertices[triangle.vertices[i]].index, triangle.polygon_index)
                self.calc_colors[key] = blue if uv_normal.z >= 0.0 else red

    def recalculate_uv_overlap(self, obj, mesh):
        self.calc_colors = dict()
        blue = 2.0 / 3.0 #blue
        no_color = -1.0
        for _, triangle in enumerate(mesh.loop_triangles):
            polygon = mesh.polygons[triangle.polygon_index]
            if polygon.hide and obj.mode == 'EDIT':
                continue
            
            for i in range(3):
                loop_index = triangle.loops[i]
                uv = mesh.uv_layers.active.data[loop_index]
                key = (mesh.vertices[triangle.vertices[i]].index, triangle.polygon_index)
                self.calc_colors[key] = blue if uv.select else no_color


    # 3D tri area ABC is half the length of AB cross product AC 
    def tri_area(self, co1, co2, co3 ):
        return (co2 - co1).cross( co3 - co1 ).length / 2.0

    def get_angles(self, vert1: Vector, vert2: Vector, vert3: Vector):
        edge1 = vert2 - vert1
        edge2 = vert3 - vert1
        angle1 = edge1.angle(edge2)

        edge1 = vert3 - vert2
        edge2 = vert1 - vert2
        angle2 = edge1.angle(edge2)

        edge1 = vert1 - vert3
        edge2 = vert2 - vert3
        angle3 = edge1.angle(edge2)

        return angle1, angle2, angle3
    

    def recalculate_stretching(self, obj, mesh):
        if self.invoked_obj.uv_stretch_type == 'ANGLES':
            self.recalculate_angle_stretching(obj, mesh)
        else:
            self.recalculate_area_stretching(obj, mesh)


    def recalculate_area_stretching(self, obj, mesh):
        self.calc_colors = dict()
        triangle_coords = [(0.0, 0.0, 0.0) for i in range(3)]
        uv_coords = [(0.0, 0.0, 0.0) for i in range(3)]

        for _, triangle in enumerate(mesh.loop_triangles):
            if mesh.polygons[triangle.polygon_index].hide and obj.mode == 'EDIT':
                continue

            for i in range(3):
                loop_index = triangle.loops[i]
                uv_coords[i] = Vector((*mesh.uv_layers.active.data[loop_index].uv, 0))
                triangle_coords[i] = Vector(mesh.vertices[triangle.vertices[i]].co)
            triangle_vec = Vector((v1 - v0).length for v0, v1 in zip(triangle_coords, triangle_coords[1:] + triangle_coords[0:1])).normalized()
            uv_vec = Vector((v1 - v0).length for v0, v1 in zip(uv_coords, uv_coords[1:] + uv_coords[0:1])).normalized()

            difference = (triangle_vec - uv_vec).length
            max_difference = obj.uv_max_difference
            if difference > max_difference:
                difference = max_difference

            max_h = 2.0 / 3.0 #blue
            norm_h = max_h - (difference - 0.0) / (max_difference - 0.0) * max_h
            for i in range(3):
                key = (mesh.vertices[triangle.vertices[i]].index, triangle.polygon_index)
                if not self.calc_colors.get(key):
                    self.calc_colors[key] = norm_h
                else:
                    self.calc_colors[key] = (self.calc_colors[key] + norm_h) / 2

    def recalculate_angle_stretching(self, obj, mesh):
        self.calc_colors = dict()
        triangle_coords = [(0.0, 0.0, 0.0) for i in range(3)]
        uv_coords = [(0.0, 0.0, 0.0) for i in range(3)]

        for _, triangle in enumerate(mesh.loop_triangles):
            if mesh.polygons[triangle.polygon_index].hide and obj.mode == 'EDIT':
                continue

            for i in range(3):
                loop_index = triangle.loops[i]
                uv_coords[i] = Vector((*mesh.uv_layers.active.data[loop_index].uv, 0))
                triangle_coords[i] = Vector(mesh.vertices[triangle.vertices[i]].co)
            tri_angles = self.get_angles(*triangle_coords)
            uv_angles = self.get_angles(*uv_coords)

            for i in range(3):
                tri_angle = tri_angles[i]
                uv_angle = uv_angles[i]
                division = tri_angle / uv_angle if tri_angle > uv_angle else uv_angle / tri_angle
                max_division = obj.uv_max_division
                if division > max_division:
                    division = max_division

                max_h = 2.0 / 3.0 #blue
                norm_h = max_h - (division - 1.0) / (max_division - 1.0) * max_h
                key = (mesh.vertices[triangle.vertices[i]].index, triangle.polygon_index)
                if not self.calc_colors.get(key):
                    self.calc_colors[key] = norm_h
                else:
                    prev = self.calc_colors[key]
                    self.calc_colors[key] = (prev + norm_h) / 2
            
    def recalculate_vertices(self, context, obj):
        self.verts = []
        self.uvs = []
        self.vert_directions = []
        self.triangle_indices = []
        self.uv_indices = []
        self.colors = []
        self.uv_colors = []
        self.edge_indices = []
        self.vert_idx_cache = set()
        self.tex_coords = []
        self.coords = []
        self.outline_coords = []
        self.wireframe_directions = []
        self.wireframe_vertex_colors = []
        self.wireframe_edge_colors = []
        self.outline_colors = []
        
        if context.window_manager.uv_select_overlap and obj.mode == 'EDIT':
            context.window_manager.uv_select_overlap = False
            bpy.ops.uv.select_overlap()
            for ob in context.selected_objects:
                ob.uv_overlap_recalculate = True
            
        context.view_layer.update()
        obj.update_from_editmode()
        depsgraph = context.evaluated_depsgraph_get()
        obj = obj.evaluated_get(depsgraph)
        mesh = obj.to_mesh(preserve_all_data_layers=True, depsgraph=depsgraph)

        if self.invoked_obj.uv_operation == 'UV_ISLANDS' or self.invoked_obj.uv_enable_explosion_view:
            self.recalculate_poly_islands(context, mesh)

        mesh.calc_loop_triangles()

        
        if self.invoked_obj.uv_operation == 'UV_STRETCHING':
            self.recalculate_stretching(obj, mesh)
        elif self.invoked_obj.uv_operation == 'UV_NORMALS':
            self.recalculate_uv_normals(obj, mesh)
        elif self.invoked_obj.uv_operation == 'UV_OVERLAP':
            if self.invoked_obj.uv_overlap_recalculate:
                self.invoked_obj.uv_overlap_recalculate = False
                self.recalculate_uv_overlap(obj, mesh)

        idx = 0
        uv_idx = 0
        for _, triangle in enumerate(mesh.loop_triangles):
            polygon = mesh.polygons[triangle.polygon_index]
            if polygon.hide and obj.mode == 'EDIT':
                continue
            
            # calculate wireframe
            if (self.invoked_obj.uv_show_wire and self.invoked_obj.mode == 'EDIT') \
                or (self.invoked_obj.select_get() and self.invoked_obj.mode == 'OBJECT'):
                # offset wireframe verts to have slightly different depth - works well only for close view range
                for loop_index in polygon.loop_indices:
                    loop = mesh.loops[loop_index]
                    edge = mesh.edges[loop.edge_index]
                    for vert_index in edge.vertices:
                        vertex = mesh.vertices[vert_index]
                        self.coords.append(vertex.co + vertex.normal * 0.0035)

                        if self.invoked_obj.uv_enable_explosion_view:
                            direction = self.directions[polygon.index][:]
                        else:
                            direction = (0.0, 0.0, 0.0)
                        self.wireframe_directions.append(direction)

                        theme_colors = context.preferences.themes["Default"].view_3d
                        
                        if self.invoked_obj.uv_show_wire and self.invoked_obj.mode == 'EDIT':
                            if edge.select:
                                edge_color = theme_colors.edge_select
                            elif edge.use_seam:
                                edge_color = theme_colors.edge_seam
                            else:
                                edge_color = theme_colors.wire_edit
                            self.wireframe_edge_colors.append(edge_color)

                            if vertex.select:
                                vertex_color = theme_colors.vertex_select
                            else:
                                vertex_color = theme_colors.vertex
                            self.wireframe_vertex_colors.append(vertex_color)

                        if self.invoked_obj.select_get() and self.invoked_obj.mode == 'OBJECT':
                            if self.invoked_obj is context.object:
                                outline_color = theme_colors.object_active
                            else:
                                outline_color = theme_colors.object_selected
                            self.outline_colors.append(outline_color)
                            self.outline_coords.append(vertex.co + vertex.normal * -0.0035)


            for i in range(3):
                vertex = mesh.vertices[triangle.vertices[i]]
                if self.invoked_obj.uv_operation == 'NONE':
                    hue = 0
                elif self.invoked_obj.uv_operation == 'UV_ISLANDS':
                    hue = self.island_colors[triangle.polygon_index]
                else:
                    hue = self.calc_colors[(vertex.index, triangle.polygon_index)]

                loop_index = triangle.loops[i]
                vertex_co = vertex.co + vertex.normal * 0.0035

                if self.invoked_obj.uv_enable_explosion_view:
                    direction = self.directions[triangle.polygon_index][:]
                else:
                    direction = (0.0, 0.0, 0.0)

                self.verts.append(vertex_co)
                self.vert_directions.append(direction)
                self.colors.append(hue)
                self.tex_coords.append(mesh.uv_layers.active.data[loop_index].uv)
                
                if polygon.select:
                    self.uvs.append(mesh.uv_layers.active.data[loop_index].uv)
                    self.uv_colors.append(hue)

            if polygon.select:
                self.uv_indices.append([uv_idx, uv_idx+1, uv_idx+2])
                uv_idx = uv_idx + 3

            self.triangle_indices.append([idx, idx+1, idx+2])
            idx = idx + 3
        print()    
                

    def draw_overlay_uv(self):
        context = bpy.context
        obj = self.invoked_obj
        try:
            if not obj or not obj.rv_enabled:
                return {'FINISHED'}
        except ReferenceError:
            return {'FINISHED'}

        if context.space_data.mode != 'UV':
            return
        if not obj.uv_show_2D:
            return
        if not obj.select_get():
            return
        if obj.mode != 'EDIT':
            return

        vertices = ((0,1),(1,1),
                        (0,0),(1,0))
        indices = ((0,1,2),(2,1,3))
        if not obj.uv_fill_texture:
            vertices = self.uvs
            indices = self.uv_indices
        
        # This shader has to be created here ALWAYS! Do not nest it!
        # Why? I have no idea, but without it, visualization of UVs of multiple objects doesn't work.
        texture_shader = gpu.shader.create_from_info(shader_info_texture)
        texture_batch = batch_for_shader(
            texture_shader, 'TRIS',
            {"position": vertices,},
            indices=indices,
        )

        texture_shader.uniform_sampler("image", self.texture)

        # Create float buffer with padding => final size has to be multiple of vec4
        buf_tex2d = gpu.types.Buffer('FLOAT', 4, (obj.uv_texture_scale, obj.uv_texture_alpha, 0, 0))
        ubo_tex2d = gpu.types.GPUUniformBuf(buf_tex2d)
        texture_shader.uniform_block("ubo_tex2d", ubo_tex2d)

        if obj.uv_checker_texture and obj is bpy.context.object:
            gpu.state.depth_test_set('LESS_EQUAL')
            gpu.state.blend_set('ALPHA')

            texture_batch.draw(texture_shader)

            gpu.state.depth_test_set('NONE')
            gpu.state.blend_set('NONE')
        
        if obj.uv_checker_texture and obj.uv_operation == 'NONE':
            return

        shader = gpu.shader.create_from_info(shader_info_2D)
        batch = batch_for_shader(
            shader, 'TRIS',
            {"position": self.uvs, "hue": self.uv_colors, },
            indices=self.uv_indices,
        )

        gpu.state.depth_test_set('LESS_EQUAL')
        gpu.state.blend_set('MULTIPLY' if obj.uv_checker_texture else 'ALPHA')

        hue_shift = obj.uv_hue_shift if obj.uv_enable_color_change else 0.0
        saturation = obj.uv_saturation if obj.uv_enable_color_change else 1.0
        value = obj.uv_value if obj.uv_enable_color_change else 1.0
        alpha = obj.rv_groups_alpha

        # Create float buffer with padding => final size has to be multiple of vec4
        buf_color = gpu.types.Buffer('FLOAT', 4, (hue_shift, saturation, value, alpha))
        ubo_color = gpu.types.GPUUniformBuf(buf_color)
        shader.uniform_block("ubo_color", ubo_color)
        
        batch.draw(shader)

        gpu.state.depth_test_set('NONE')
        gpu.state.blend_set('NONE')


    def draw_overlay(self):
        obj = self.invoked_obj
        try:
            if not obj or not obj.rv_enabled:
                return {'FINISHED'}
        except ReferenceError:
            return {'FINISHED'}
        
        if not obj.uv_show_3D:
            return

        shader = gpu.shader.create_from_info(shader_info)
        batch = batch_for_shader(
            shader, 'TRIS',
            {"position": self.verts, "hue": self.colors, "direction": self.vert_directions, "uv": self.tex_coords},
            indices=self.triangle_indices,
        )
        

        if obj.rv_backface_culling:
            gpu.state.face_culling_set('BACK')
        
        gpu.state.depth_test_set('LESS_EQUAL')

        gpu.state.blend_set('NONE')

        for area in bpy.context.screen.areas:
            if area.type == 'VIEW_3D':
                for space in area.spaces:
                    if space.type == 'VIEW_3D':
                        if space.shading.type == 'WIREFRAME':
                            gpu.state.depth_test_set('ALWAYS')

        if obj.show_in_front:
            gpu.state.depth_test_set('ALWAYS')
            gpu.state.face_culling_set('BACK')

        shader.uniform_float("viewProjectionMatrix", bpy.context.region_data.perspective_matrix)
        shader.uniform_float("worldMatrix", obj.matrix_world)

        hue_shift = obj.uv_hue_shift if obj.uv_enable_color_change else 0.0
        saturation = obj.uv_saturation if obj.uv_enable_color_change else 1.0
        value = obj.uv_value if obj.uv_enable_color_change else 1.0
        alpha = obj.rv_groups_alpha

        scale_factor = obj.uv_texture_scale
        tex_enabled = 1 if obj.uv_checker_texture else 0
        tex_only = 1 if obj.uv_checker_texture and obj.uv_operation == 'NONE' else 0

        x_offset = obj.uv_location_offset[0] if obj.uv_enable_position_change else 0.0
        y_offset = obj.uv_location_offset[1] if obj.uv_enable_position_change else 0.0
        z_offset = obj.uv_location_offset[2] if obj.uv_enable_position_change else 0.0
        explosion_offset = obj.uv_explosion_offset if obj.uv_enable_explosion_view else 0.0

        # Create float buffer with padding => final size has to be multiple of vec4
        buf_3d = gpu.types.Buffer('FLOAT', 12, (hue_shift, saturation, value, alpha, \
                                                scale_factor, tex_enabled, tex_only, \
                                                x_offset, y_offset, z_offset, explosion_offset, 0))
        ubo_3d = gpu.types.GPUUniformBuf(buf_3d)
        shader.uniform_block("ubo_3d", ubo_3d)

        shader.uniform_sampler("image", self.texture)

        batch.draw(shader)

        # draw outline - not finished
        if obj.select_get() and obj.mode == 'OBJECT':
            gpu.state.line_width_set(20)
            gpu.state.depth_test_set('GREATER')
            gpu.state.blend_set('NONE')
            outline_shader = gpu.shader.create_from_info(shader_info_wireframe)
            outline_batch = batch_for_shader(outline_shader, 'LINES', {"position": self.outline_coords,
                                                                            "direction": self.wireframe_directions,
                                                                            "color": self.outline_colors})

            outline_shader.uniform_float("viewProjectionMatrix", bpy.context.region_data.perspective_matrix)
            outline_shader.uniform_float("worldMatrix", obj.matrix_world)

            # Create float buffer with padding => final size has to be multiple of vec4
            buf_outline = gpu.types.Buffer('FLOAT', 4, (x_offset, y_offset, z_offset, explosion_offset))
            ubo_outline = gpu.types.GPUUniformBuf(buf_outline)
            outline_shader.uniform_block("ubo_wire", ubo_outline)

            #outline_batch.draw(outline_shader)

        # draw wireframe
        if obj.mode == 'EDIT' and obj.uv_show_wire:
            gpu.state.line_width_set(2)
            gpu.state.depth_test_set('LESS_EQUAL')
            gpu.state.blend_set('NONE')
            wireframe_shader = gpu.shader.create_from_info(shader_info_wireframe)
            wireframe_edge_batch = batch_for_shader(wireframe_shader, 'LINES', {"position": self.coords,
                                                                            "direction": self.wireframe_directions,
                                                                            "color": self.wireframe_edge_colors})

            wireframe_shader.uniform_float("viewProjectionMatrix", bpy.context.region_data.perspective_matrix)
            wireframe_shader.uniform_float("worldMatrix", obj.matrix_world)

            # Create float buffer with padding => final size has to be multiple of vec4
            buf_wire = gpu.types.Buffer('FLOAT', 4, (x_offset, y_offset, z_offset, explosion_offset))
            ubo_wire = gpu.types.GPUUniformBuf(buf_wire)
            wireframe_shader.uniform_block("ubo_wire", ubo_wire)

            wireframe_edge_batch.draw(wireframe_shader)

            wireframe_vertex_batch = batch_for_shader(wireframe_shader, 'POINTS', {"position": self.coords,
                                                                            "direction": self.wireframe_directions,
                                                                            "color": self.wireframe_vertex_colors})
            gpu.state.point_size_set(6)

            if bpy.context.tool_settings.mesh_select_mode[0]:
                wireframe_vertex_batch.draw(wireframe_shader)

        gpu.state.point_size_set(1)
        gpu.state.line_width_set(1)
        gpu.state.depth_test_set('NONE')
        gpu.state.blend_set('NONE')
        
        if obj.rv_backface_culling:
            gpu.state.face_culling_set('NONE')
        
        #print("DRAWING OVERLAY!!!")


    def modal(self, context, event):
        try:
            if not self.invoked_obj.rv_enabled:
                bpy.types.SpaceImageEditor.draw_handler_remove(self.overlay_handler_2D, 'WINDOW')
                bpy.types.SpaceView3D.draw_handler_remove(self.overlay_handler_3D, 'WINDOW')
                bpy.data.images.remove(self.chess_image)
                return {'FINISHED'}
        except ReferenceError:
            bpy.types.SpaceImageEditor.draw_handler_remove(self.overlay_handler_2D, 'WINDOW')
            bpy.types.SpaceView3D.draw_handler_remove(self.overlay_handler_3D, 'WINDOW')
            bpy.data.images.remove(self.chess_image)
            return {'FINISHED'}

        for area in bpy.context.screen.areas :
            if area.type == 'IMAGE_EDITOR' or area.type == 'VIEW_3D':
                area.tag_redraw()

        
        #if event:
            #print(event.type + " " + event.value)
        #print(self.current_stretch_type + " " + self.invoked_obj.uv_stretch_type)

        self.check_image_exists()
        self.texture = gpu.texture.from_image(self.invoked_obj.uv_image)
        
        if self.current_stretch_type != self.invoked_obj.uv_stretch_type:
            self.current_stretch_type = self.invoked_obj.uv_stretch_type
            self.invoked_obj.uv_recalculate = True
            return {'PASS_THROUGH'}
        elif event and (event.value in ('RELEASE', 'CLICK')):
            self.invoked_obj.uv_recalculate = True
            return {'PASS_THROUGH'}

        if self.invoked_obj.uv_recalculate or self.invoked_obj.uv_realtime:
            self.recalculate_vertices(context, self.invoked_obj)
            self.invoked_obj.uv_recalculate = False

        return {'PASS_THROUGH'}

    def invoke(self, context, event):
        args = ()
        obj = context.object
        self.event_type = event.type
        self.invoked_obj = context.object
        self.create_image()
        self.invoked_obj.uv_image = self.chess_image
        self.texture = gpu.texture.from_image(self.invoked_obj.uv_image)
        self.current_stretch_type = self.invoked_obj.uv_stretch_type

        self.verts = []
        self.uvs = []
        self.vert_directions = []
        self.triangle_indices = []
        self.uv_indices = []
        self.colors = []
        self.uv_colors = []
        self.edge_indices = []
        self.vert_idx_cache = set()
        self.tex_coords = []

        if self.invoked_obj.uv_operation == 'UV_OVERLAP':
                context.window_manager.uv_select_overlap = True

        self.recalculate_vertices(context, obj)
        self.overlay_handler_3D = bpy.types.SpaceView3D.draw_handler_add(self.draw_overlay, args, 'WINDOW', 'POST_VIEW')
        self.overlay_handler_2D = bpy.types.SpaceImageEditor.draw_handler_add(self.draw_overlay_uv, (), 'WINDOW', 'POST_VIEW')
        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}

    def create_image(self):
        script_file = os.path.realpath(__file__)
        directory = os.path.dirname(script_file)                
        self.chess_image = bpy.data.images.load(filepath= os.path.join(directory, "tiles.png"))

    def check_image_exists(self):
        if not self.chess_image:
            self.create_image()
        if not self.invoked_obj.uv_image:
            self.invoked_obj.uv_image = self.chess_image